/*
 * SPDX-License-Identifier: GPL-3.0-or-later
 * SPDX-FileCopyrightText: 2020 Johan Ouwerkerk <jm.ouwerkerk@gmail.com>
 */
#ifndef SECRET_SODIUM_H
#define SECRET_SODIUM_H

extern "C" {
#include <sodium.h>
}

#endif
